#include "depth_shader_class.h"

#include <d3dcompiler.h>

#include "logger.h"

depth_shader_class::depth_shader_class()
{
    vertex_shader_ = nullptr;
    pixel_shader_ = nullptr;
    layout_ = nullptr;
    matrix_buffer_ = nullptr;
}

depth_shader_class::~depth_shader_class()
{
    shutdown();
}

bool depth_shader_class::initialize(ID3D11Device* device, HWND hwnd)
{

    Logger::Get().Log("Initializing depth shader class", __FILE__, __LINE__, Logger::LogLevel::Initialize);
    
    // Charger et compiler les shaders (remplacez les chemins par les vtres)
    bool result = initialize_shader(device, hwnd, L"src/hlsl/depth.vs", L"src/hlsl/depth.ps");
    if (!result)
    {
        Logger::Get().Log("Failed to initialize depth shader", __FILE__, __LINE__, Logger::LogLevel::Error);
        return false;
    }

    Logger::Get().Log("Initialized depth shader class", __FILE__, __LINE__, Logger::LogLevel::Info);
    
    return true;
}

void depth_shader_class::shutdown()
{
    shutdown_shader();
}

bool depth_shader_class::render(ID3D11DeviceContext* deviceContext, int indexCount, XMMATRIX worldMatrix, XMMATRIX viewMatrix, XMMATRIX projectionMatrix)
{
    if (!set_shader_parameters(deviceContext, worldMatrix, viewMatrix, projectionMatrix))
        return false;

    render_shader(deviceContext, indexCount);
    return true;
}

bool depth_shader_class::initialize_shader(ID3D11Device* device, HWND hwnd, const WCHAR* vsFilename, const WCHAR* psFilename)
{
    HRESULT result;
    ID3DBlob* vertexShaderBuffer = nullptr;
    ID3DBlob* pixelShaderBuffer = nullptr;
    ID3DBlob* errorMessage = nullptr;

    // Compiler le vertex shader
    result = D3DCompileFromFile(vsFilename, nullptr, nullptr, "DepthVertexShader", "vs_5_0", 0, 0, &vertexShaderBuffer, &errorMessage);
    if (FAILED(result))
    {
        if (errorMessage)
        {
            output_shader_error_message(errorMessage, hwnd, vsFilename);
            errorMessage->Release();
        }
        else
        {
            Logger::Get().Log("Fichier de vertex shader introuvable: " + std::to_string(reinterpret_cast<uintptr_t>(vsFilename)), __FILE__, __LINE__, Logger::LogLevel::Error);
        }
        return false;
    }

    // Compiler le pixel shader
    errorMessage = nullptr;
    result = D3DCompileFromFile(psFilename, nullptr, nullptr, "DepthPixelShader", "ps_5_0", 0, 0, &pixelShaderBuffer, &errorMessage);
    if (FAILED(result))
    {
        if (errorMessage)
        {
            output_shader_error_message(errorMessage, hwnd, psFilename);
            errorMessage->Release();
        }
        else
        {
            Logger::Get().Log("Fichier de pixel shader introuvable: " + std::to_string(reinterpret_cast<uintptr_t>(psFilename)), __FILE__, __LINE__, Logger::LogLevel::Error);
        }
        
        if (vertexShaderBuffer)
        {
            vertexShaderBuffer->Release();
        }
        return false;
    }

    // Crer les shaders
    device->CreateVertexShader(vertexShaderBuffer->GetBufferPointer(), vertexShaderBuffer->GetBufferSize(), nullptr, &vertex_shader_);
    device->CreatePixelShader(pixelShaderBuffer->GetBufferPointer(), pixelShaderBuffer->GetBufferSize(), nullptr, &pixel_shader_);

    // Dfinir la description du layout d'entre
    D3D11_INPUT_ELEMENT_DESC polygonLayout[1];
    polygonLayout[0].SemanticName = "POSITION";
    polygonLayout[0].SemanticIndex = 0;
    polygonLayout[0].Format = DXGI_FORMAT_R32G32B32_FLOAT;
    polygonLayout[0].InputSlot = 0;
    polygonLayout[0].AlignedByteOffset = 0;
    polygonLayout[0].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
    polygonLayout[0].InstanceDataStepRate = 0;

    // Crer le layout
    device->CreateInputLayout(polygonLayout, 1, vertexShaderBuffer->GetBufferPointer(), vertexShaderBuffer->GetBufferSize(), &layout_);

    // Librer les buffers
    vertexShaderBuffer->Release();
    pixelShaderBuffer->Release();

    // Crer le buffer de matrices
    D3D11_BUFFER_DESC matrixBufferDesc = {};
    matrixBufferDesc.Usage = D3D11_USAGE_DYNAMIC;
    matrixBufferDesc.ByteWidth = sizeof(MatrixBufferType);
    matrixBufferDesc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    matrixBufferDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
    device->CreateBuffer(&matrixBufferDesc, nullptr, &matrix_buffer_);

    return true;
}

bool depth_shader_class::set_shader_parameters(ID3D11DeviceContext* deviceContext, XMMATRIX worldMatrix, XMMATRIX viewMatrix, XMMATRIX projectionMatrix)
{
    D3D11_MAPPED_SUBRESOURCE mappedResource;
    MatrixBufferType* dataPtr;

    // Transposer les matrices pour le shader
    worldMatrix = XMMatrixTranspose(worldMatrix);
    viewMatrix = XMMatrixTranspose(viewMatrix);
    projectionMatrix = XMMatrixTranspose(projectionMatrix);

    // Mapper le buffer
    HRESULT result = deviceContext->Map(matrix_buffer_, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
    if (FAILED(result))
        return false;

    dataPtr = (MatrixBufferType*)mappedResource.pData;
    dataPtr->world = worldMatrix;
    dataPtr->view = viewMatrix;
    dataPtr->projection = projectionMatrix;

    deviceContext->Unmap(matrix_buffer_, 0);

    // Dfinir le buffer de constantes pour le vertex shader
    deviceContext->VSSetConstantBuffers(0, 1, &matrix_buffer_);

    return true;
}

void depth_shader_class::render_shader(ID3D11DeviceContext* deviceContext, int indexCount)
{
    // Dfinir le layout d'entre
    deviceContext->IASetInputLayout(layout_);

    // Dfinir les shaders
    deviceContext->VSSetShader(vertex_shader_, nullptr, 0);
    deviceContext->PSSetShader(pixel_shader_, nullptr, 0);

    // Dessiner
    deviceContext->DrawIndexed(indexCount, 0, 0);
}

void depth_shader_class::shutdown_shader()
{
    // Librer le buffer de matrice
    if (matrix_buffer_)
    {
        matrix_buffer_->Release();
        matrix_buffer_ = nullptr;
    }

    // Librer le layout
    if (layout_)
    {
        layout_->Release();
        layout_ = nullptr;
    }

    // Librer le pixel shader
    if (pixel_shader_)
    {
        pixel_shader_->Release();
        pixel_shader_ = nullptr;
    }

    // Librer le vertex shader
    if (vertex_shader_)
    {
        vertex_shader_->Release();
        vertex_shader_ = nullptr;
    }
}

void depth_shader_class::output_shader_error_message(ID3DBlob* errorMessage, HWND hwnd, const WCHAR* shaderFilename)
{
    // Obtenir un pointeur vers le message d'erreur
    char* compileErrors = (char*)(errorMessage->GetBufferPointer());

    // Obtenir la taille du message
    SIZE_T bufferSize = errorMessage->GetBufferSize();
    
    // Allouer de la mmoire pour le message d'erreur format
    char* message = new char[bufferSize + 100];
    
    // Dfinir le message d'erreur format
    sprintf_s(message, bufferSize + 100, "Erreur de compilation du shader %ls :\n%s", 
              shaderFilename, compileErrors);
    
    // Afficher le message d'erreur dans la console
    Logger::Get().Log(message, __FILE__, __LINE__, Logger::LogLevel::Error);
    
    // Afficher une bote de dialogue avec l'erreur
    MessageBoxA(hwnd, message, "Erreur de Shader", MB_OK);
    
    // Librer la mmoire
    delete[] message;
}